@@ -1,3 +1,44 @@
+"""
+Argument definition
+
+This module enables the definitions of various kinds of configuration arguments.
+
+- Parameters (:class:`.Param`) correspond to values that will be present in the configuration. 
+  Each invocation of a parameter has a user-provided string attached, which is parsed by a 
+  :class:`~configpile.types.ParamType`. Those invocations can come from environment variables
+  (:attr:`.Param.env_var_name`), configuration files (:attr:`.Param.config_key_name`),
+  command-line flags (:attr:`.Arg.short_flag_name`, :attr:`.Arg.long_flag_name`), or finally
+  as positional command-line parameters (:attr:`.Param.positional`).
+
+- Configuration file inclusion using special parameters :meth:`.Param.config`
+
+- Expanders (:class:`.Expander`) can only be present as command-line flags; their action is
+  to insert a key/value pair in the command-line. This enables, for example, flags that set
+  a boolean parameter to `True` or `False`. 
+
+.. rubric:: Types
+
+This module uses the following types. 
+
+.. py:data:: _Config
+
+    Configuration dataclass type being constructed
+
+.. py:data:: _Arg
+
+    Precise argument type, used mostly internally.
+
+    See `<https://en.wikipedia.org/wiki/Bounded_quantification>`_
+
+.. py:data:: _Value
+
+    Type of the parameter value
+
+.. py:data:: _Item
+
+    Item in a sequence
+"""
+
 from __future__ import annotations
 
 import dataclasses
@@ -8,7 +49,6 @@ from pathlib import Path
 from typing import (
     TYPE_CHECKING,
     Any,
-    Callable,
     Dict,
     Generic,
     List,
@@ -20,89 +60,54 @@ from typing import (
     Union,
 )
 
-from typing_extensions import TypeGuard
-
 from .collector import Collector
-from .errors import Err
-from .types import ParamType, path
+from .parsers import Parser, path_parser
 
-T = TypeVar("T", covariant=True)  #: Item type
+# documented in the module docstring
 
-W = TypeVar("W", covariant=True)  #: Wrapped item type
+_Value = TypeVar("_Value", covariant=True)
+
+_Item = TypeVar("_Item", covariant=True)
 
 if TYPE_CHECKING:
     from .config import Config
     from .processor import ProcessorFactory
 
 
-class AutoName(Enum):
+class Derived(Enum):
     """
     Describes automatic handling of an argument name
     """
 
-    #: The argument should not be present in the corresponding source
-    FORBIDDEN = 0
+    #: Derives the name/key using snake_case
+    SNAKE_CASE = 1
 
-    #: Derives the argument name from the original Python identifier (default)
-    DERIVED = 1
+    #: Derives the name/key using SNAKE_CASE (and sets uppercase, default for env. variables)
+    SNAKE_CASE_UPPER_CASE = 2
 
-    @staticmethod
-    def derive_long_flag_name(name: str) -> str:
+    #: Derives the argument name using kebab-case (default for INI file keys and cmd line flags)
+    KEBAB_CASE = 3
+
+    def derive(self, name: str) -> str:
         """
-        Returns a long flag name
-
-        Changes the snake_case to kebab-case and adds a ``--`` prefix
+        Returns a derived name from a field name
 
         Args:
-            name: Python identifier used to derive the flag
-
-        Returns:
-            A long flag name
+            name: Field name in Python identifier format
         """
-        if name.endswith("_command"):
-            name = name[:-8]
-        return "--" + name.replace("_", "-")
-
-    @staticmethod
-    def derive_env_var_name(name: str, prefix: Optional[str]) -> str:
-        """
-        Returns a environment variable name derived from a Python identifier
-
-        Keeps snake_case but transforms into upper case, and optionally adds a prefix,
-        separated by an underscore.
-
-        Args:
-            name: Python identifier used to derive the flag
-            prefix: Optional prefix
-
-        Returns:
-            An environment variable name
-        """
-        if prefix is not None:
-            return prefix + "_" + name.upper()
-        else:
+        assert str.isidentifier(name)
+        if self == Derived.SNAKE_CASE:
+            return name
+        elif self == Derived.SNAKE_CASE_UPPER_CASE:
             return name.upper()
-
-    @staticmethod
-    def derive_config_key_name(name: str) -> str:
-        """
-        Derives a INI key name
-
-        It matches the Python identifier, with snake_case replaced by kebab-case.
-
-        Args:
-            name: Python identifier used to derive the key name
-
-        Returns:
-            INI file key name
-        """
-        return name.replace("_", "-")
+        elif self == Derived.KEBAB_CASE:
+            return name.replace("_", "-")
+        else:
+            raise NotImplementedError
 
 
-ArgName = Union[str, AutoName]
-
-A = TypeVar("A", bound="Arg")
-C = TypeVar("C", bound="Config")
+_Arg = TypeVar("_Arg", bound="Arg")
+_Config = TypeVar("_Config", bound="Config")
 
 # TODO: solve this bug
 
@@ -113,15 +118,18 @@ class Arg(ABC):
     Base class for all kinds of arguments
     """
 
-    help: Optional[str]  #: Help for the argument
+    #: Help for the argument
+    help: Optional[str]
 
     #: Short option name, used in command line parsing, prefixed by a single hyphen
     short_flag_name: Optional[str]
 
     #: Long option name used in command line argument parsing
     #:
-    #: It is lowercase, prefixed with ``--`` and words are separated by hyphens
-    long_flag_name: ArgName
+    #: It is in general lowercase, prefixed with ``--`` and words are separated by hyphens.,
+    #:
+    #: If set to
+    long_flag_name: Union[str, Derived, None]
 
     def all_flags(self) -> Sequence[str]:
         """
@@ -130,12 +138,14 @@ class Arg(ABC):
         res: List[str] = []
         if self.short_flag_name is not None:
             res.append(self.short_flag_name)
-        assert self.long_flag_name != AutoName.DERIVED
+        assert not isinstance(self.long_flag_name, Derived)
         if isinstance(self.long_flag_name, str):
             res.append(self.long_flag_name)
         return res
 
-    def update_dict_(self, name: str, help: str, env_prefix: Optional[str]) -> Mapping[str, Any]:
+    def update_dict_(
+        self, name: str, help: Optional[str], env_prefix: Optional[str]
+    ) -> Mapping[str, Any]:
         """
         Returns updated values for this argument, used during :class:`.App` construction
 
@@ -147,12 +157,14 @@ class Arg(ABC):
         Returns:
 
         """
-        res = {"help": help}
-        if self.long_flag_name == AutoName.DERIVED:
-            res["long_flag_name"] = AutoName.derive_long_flag_name(name)
+        res: Dict[str, Any] = {}
+        if self.help is None:
+            res = {"help": help}
+        if isinstance(self.long_flag_name, Derived):
+            res["long_flag_name"] = "--" + self.long_flag_name.derive(name)
         return res
 
-    def updated(self: A, name: str, help: str, env_prefix: Optional[str]) -> A:
+    def updated(self: _Arg, name: str, help: str, env_prefix: Optional[str]) -> _Arg:
         """
         Returns a copy of this argument with some data updated from its declaration context
 
@@ -168,7 +180,7 @@ class Arg(ABC):
         return dataclasses.replace(self, **self.update_dict_(name, help, env_prefix))
 
     @abstractmethod
-    def update_processor(self, pf: ProcessorFactory[C]) -> None:
+    def update_processor(self, pf: ProcessorFactory[_Config]) -> None:
         """
         Updates a config processor with the processing required by this argument
 
@@ -195,6 +207,7 @@ class Expander(Arg):
     """
 
     new_flag: str  #: Inserted flag in the command line
+
     new_value: str  #: Inserted value in the command line
 
     def inserts(self) -> Tuple[str, str]:
@@ -203,7 +216,7 @@ class Expander(Arg):
         """
         return (self.new_flag, self.new_value)
 
-    def update_processor(self, pf: ProcessorFactory[C]) -> None:
+    def update_processor(self, pf: ProcessorFactory[_Config]) -> None:
         from .processor import CLInserter
 
         for flag in self.all_flags():
@@ -217,7 +230,7 @@ class Expander(Arg):
         *,
         help: Optional[str] = None,
         short_flag_name: Optional[str] = None,
-        long_flag_name: ArgName = AutoName.DERIVED,
+        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
     ) -> Expander:
         """
         Constructs an expander that inserts a flag/value pair in the command line
@@ -251,10 +264,9 @@ class Positional(Enum):
     Describes the positional behavior of a parameter
     """
 
-    FORBIDDEN = 0  #: The argument is not positional
-    ONCE = 1  #: The argument parses a single positional value
-    ZERO_OR_MORE = 2  #: The argument parses the remaining positional value
-    ONE_OR_MORE = 3  #: The argument parses at least one remaining positional value
+    ONCE = 1  #: Eats a single positional value
+    ZERO_OR_MORE = 2  #: Eats as many positional values as possible
+    ONE_OR_MORE = 3  #: Eats as many positional values as possible, but at least one
 
     def should_be_last(self) -> bool:
         """
@@ -262,68 +274,66 @@ class Positional(Enum):
         """
         return self in {Positional.ZERO_OR_MORE, Positional.ONE_OR_MORE}
 
-    def is_positional(self) -> bool:
-        """
-        Returns whether a parameter is positional
-        """
-        return self != Positional.FORBIDDEN
-
 
 @dataclass(frozen=True)
-class Param(Arg, Generic[T]):
+class Param(Arg, Generic[_Value]):
     """
     Describes an argument holding a value of a given type
 
-    .. note::
-        Instances of :class:`.Param` have two "states":
+    You'll want to construct parameters using the static methods available:
 
-        * Initially, instances of :class:`.Param` are assigned to class attributes of
-        subclasses of :class:`.app.App`. In that state, :attr:`.Param.name` is not set,
-        and the other ``XXX_name`` attributes contain either a custom name provided by the user, or
-        instructions about the derivation of the corresponding name.
+    - :meth:`~.Param.store` to create a parameter that keeps the last value provided
 
-        * When an instance of :class:`.App` is constructed, the :attr:`.name` attribute and the
-          ``XXX_name`` attributes of the instance are updated.
+    - :meth:`~.Param.append` to create a parameter that collects all the provided value into
+      a sequence. Note that the parsed value must already be a sequence; for that, use the
+      method :meth:`~configpile.types.ParamType.as_sequence_of_one` on the
+      :class:`~configpile.types.ParamType` instance.
+
+    For the two constructions above, you can construct positional parameters by setting the
+    :attr:`.Param.positional` field.
+
+    - :meth:`~.Param.config` returns a parameter that parses configuration files.
     """
 
     #: Argument type, parser from string to value
-    param_type: ParamType[T]  # type: ignore
+    parser: Parser[_Value]  # type: ignore
 
     is_config: bool  #: Whether this represent a list of config files
 
     #: Argument collector
-    collector: Collector[T]  # type: ignore
+    collector: Collector[_Value]  # type: ignore
 
     default_value: Optional[str]  #: Default value inserted as instance
 
+    #: Python field identifier for this parameter in :class:`~configpile.config.Config`
     name: Optional[str]  #: Python identifier representing the argument
 
-    positional: Positional
+    #: Whether this parameter can appear as a positional parameter and how
+    #:
+    #: A positional parameter is a paramater that appears without a preceding flag
+    positional: Optional[Positional]
 
     #: Configuration key name used in INI files
     #:
     #: It is lowercase, and words are separated by hyphens.
-    config_key_name: ArgName
+    config_key_name: Union[str, Derived, None]
 
     #: Environment variable name
     #:
-    #: The environment variable name has an optional prefix, followed by the
+    #: The environment variable name has a prefix, followed by the
     #: Python identifier in uppercase, with underscore as separator.
     #:
     #: This prefix is provided by :attr:`.App.env_prefix_`
-    #:
-    #: If a non-empty prefix is given, the name is prefixed with it
-    #: (and an underscore).
-    env_var_name: ArgName
+    env_var_name: Union[str, Derived, None]
 
-    validator: Optional[Callable[[T], Optional[Err]]]
-
-    def update_dict_(self, name: str, help: str, env_prefix: Optional[str]) -> Mapping[str, Any]:
+    def update_dict_(
+        self, name: str, help: Optional[str], env_prefix: Optional[str]
+    ) -> Mapping[str, Any]:
         r = {"name": name, **super().update_dict_(name, help, env_prefix)}
-        if self.config_key_name == AutoName.DERIVED:
-            r["config_key_name"] = AutoName.derive_config_key_name(name)
-        if self.env_var_name == AutoName.DERIVED and env_prefix is not None:
-            r["env_var_name"] = AutoName.derive_env_var_name(name, env_prefix)
+        if isinstance(self.config_key_name, Derived):
+            r["config_key_name"] = self.config_key_name.derive(name)
+        if isinstance(self.env_var_name, Derived) and env_prefix is not None:
+            r["env_var_name"] = env_prefix + self.env_var_name.derive(name)
         return r
 
     def all_config_key_names(self) -> Sequence[str]:
@@ -358,20 +368,20 @@ class Param(Arg, Generic[T]):
 
     def argparse_argument_kwargs(self) -> Mapping[str, Any]:
         res: Dict[str, Any] = {"help": self.help}
-        if self.is_required():
-            res = {**res, "required": True}
+        choices = self.parser.choices()
+        if choices is not None:
+            res = {**res, "choices": choices, "type": str}
         return {
             **res,
             **self.collector.argparse_argument_kwargs(),
-            **self.param_type.argparse_argument_kwargs(),
         }
 
-    def update_processor(self, pf: ProcessorFactory[C]) -> None:
+    def update_processor(self, pf: ProcessorFactory[_Config]) -> None:
         from .processor import CLConfigParam, CLParam, KVConfigParam, KVParam
 
         assert self.name is not None
         pf.params_by_name[self.name] = self
-        if self.positional != Positional.FORBIDDEN:
+        if self.positional is not None:
             pf.cl_positionals.append(self)
         for flag in self.all_flags():
             if self.is_config:
@@ -399,17 +409,16 @@ class Param(Arg, Generic[T]):
 
     @staticmethod
     def store(
-        param_type: ParamType[T],
+        parser: Parser[_Value],
         *,
         help: Optional[str] = None,
         default_value: Optional[str] = None,
-        positional: Positional = Positional.FORBIDDEN,
+        positional: Optional[Positional] = None,
         short_flag_name: Optional[str] = None,
-        long_flag_name: ArgName = AutoName.DERIVED,
-        config_key_name: ArgName = AutoName.DERIVED,
-        env_var_name: ArgName = AutoName.FORBIDDEN,
-        validator: Optional[Callable[[T], Optional[Err]]] = None,
-    ) -> Param[T]:
+        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
+        config_key_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
+        env_var_name: Union[str, Derived, None] = None,
+    ) -> Param[_Value]:
         """
         Creates a parameter that stores the last provided value
 
@@ -418,7 +427,7 @@ class Param(Arg, Generic[T]):
         the parameter cannot be omitted.
 
         Args:
-            param_type: Parser that transforms a string into a value
+            parser: Parser that transforms a string into a value
 
         Keyword Args:
             help: Help description (autodoc/docstring is used otherwise)
@@ -436,7 +445,7 @@ class Param(Arg, Generic[T]):
         return Param(
             name=None,
             help=help,
-            param_type=param_type,
+            parser=parser,
             collector=Collector.keep_last(),
             default_value=default_value,
             positional=positional,
@@ -445,7 +454,6 @@ class Param(Arg, Generic[T]):
             config_key_name=config_key_name,
             env_var_name=env_var_name,
             is_config=False,
-            validator=validator,
         )
 
     @staticmethod
@@ -453,9 +461,8 @@ class Param(Arg, Generic[T]):
         *,
         help: Optional[str] = None,
         short_flag_name: Optional[str] = None,
-        long_flag_name: ArgName = AutoName.DERIVED,
-        env_var_name: ArgName = AutoName.FORBIDDEN,
-        validator: Optional[Callable[[Sequence[Path]], Optional[Err]]] = None,
+        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
+        env_var_name: Union[str, Derived, None] = None,
     ) -> Param[Sequence[Path]]:
         """
         Creates a parameter that parses configuration files and stores their names
@@ -472,35 +479,36 @@ class Param(Arg, Generic[T]):
         return Param(
             name=None,
             help=help,
-            param_type=path.separated_by(",", strip=True, remove_empty=True),
+            parser=path_parser.separated_by(",", strip=True, remove_empty=True),
             collector=Collector.append(),  # type: ignore
-            positional=Positional.FORBIDDEN,
+            positional=None,
             short_flag_name=short_flag_name,
             long_flag_name=long_flag_name,
-            config_key_name=AutoName.FORBIDDEN,
+            config_key_name=None,
             env_var_name=env_var_name,
             is_config=True,
             default_value=None,
-            validator=validator,
         )
 
     @staticmethod
-    def append(
-        param_type: ParamType[Sequence[W]],
+    def append1(
+        parser: Parser[_Item],
         *,
         help: Optional[str] = None,
-        positional: Positional = Positional.FORBIDDEN,
+        positional: Optional[Positional] = None,
         short_flag_name: Optional[str] = None,
-        long_flag_name: ArgName = AutoName.DERIVED,
-        config_key_name: ArgName = AutoName.DERIVED,
-        env_var_name: ArgName = AutoName.FORBIDDEN,
-        validator: Optional[Callable[[Sequence[W]], Optional[Err]]] = None,
-    ) -> Param[Sequence[W]]:
+        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
+        config_key_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
+        env_var_name: Union[str, Derived, None] = None,
+    ) -> Param[Sequence[_Item]]:
         """
-        Creates an argument that stores the last provided value
+        Returns a newly created argument that appends single parsed values
+
+        While :meth:`~Param.append` creates a parameter that joins sequences of values, this method
+        creates a parameter that appends single items to a sequence.
 
         Args:
-            param_type: Parser that transforms a string into a value
+            parser: Parser that transforms a string into a single value item
 
         Keyword Args:
             help: Help description (autodoc/docstring is used otherwise)
@@ -509,14 +517,50 @@ class Param(Arg, Generic[T]):
             long_flag_name: Long option name (auto. derived from fieldname by default)
             config_key_name: Config key name (auto. derived from fieldname by default)
             env_var_name: Environment variable name (forbidden by default)
-
-        Returns:
-            The constructed Arg instance
         """
         return Param(
             name=None,
             help=help,
-            param_type=param_type,
+            parser=parser.as_sequence_of_one(),
+            collector=Collector.append(),  # type: ignore
+            default_value=None,
+            positional=positional,
+            short_flag_name=short_flag_name,
+            long_flag_name=long_flag_name,
+            config_key_name=config_key_name,
+            env_var_name=env_var_name,
+            is_config=False,
+        )
+
+    @staticmethod
+    def append(
+        parser: Parser[Sequence[_Item]],
+        *,
+        help: Optional[str] = None,
+        positional: Optional[Positional] = None,
+        short_flag_name: Optional[str] = None,
+        long_flag_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
+        config_key_name: Union[str, Derived, None] = Derived.KEBAB_CASE,
+        env_var_name: Union[str, Derived, None] = None,
+    ) -> Param[Sequence[_Item]]:
+        """
+        Returns a newly created argument that joins parsed sequences of values
+
+        Args:
+            parser: Parser that transforms a string into a sequence of values
+
+        Keyword Args:
+            help: Help description (autodoc/docstring is used otherwise)
+            positional: Whether this argument is present in positional arguments
+            short_flag_name: Short option name (optional)
+            long_flag_name: Long option name (auto. derived from fieldname by default)
+            config_key_name: Config key name (auto. derived from fieldname by default)
+            env_var_name: Environment variable name (forbidden by default)
+        """
+        return Param(
+            name=None,
+            help=help,
+            parser=parser,
             collector=Collector.append(),  # type: ignore
             default_value=None,
             positional=positional,
@@ -525,5 +569,4 @@ class Param(Arg, Generic[T]):
             config_key_name=config_key_name,
             env_var_name=env_var_name,
             is_config=False,
-            validator=validator,
         )
