@@ -1,8 +1,33 @@
+"""
+Arguments processing
+
+This module contains the internal machinery that processes environment variables, configuration
+files and command-line parameters.
+
+As of March 22, 2022, configpile is still pretty much influenced by :mod:`argparse`, and the
+machinery below supports a subset of :mod:`argparse` functionality. Later on, we may cut ties
+with :mod:`argparse`, add our own help/usage message writing, our own Sphinx extension and
+encourage extending those processing classes. 
+
+.. rubric:: Types
+
+This module uses the following types.
+
+.. py:data:: _Value
+
+    Value being parsed by a :class:`.ParamType`
+
+.. py:data:: _Config
+
+    Type of the configuration dataclass to construct
+"""
+
 from __future__ import annotations
 
 import argparse
 import configparser
 import sys
+import warnings
 from abc import ABC, abstractmethod
 from configparser import ConfigParser
 from dataclasses import dataclass
@@ -29,15 +54,15 @@ from typing import (
 from typing_extensions import Annotated, get_args, get_origin, get_type_hints
 
 from .arg import Arg, Expander, Param, Positional
-from .errors import Err, Result, in_context, is_err, is_value
+from .userr import Err, Res, in_context
 from .util import ClassDoc, filter_types_single
 
 if TYPE_CHECKING:
     from .config import Config
 
-C = TypeVar("C", bound="Config")
+_Config = TypeVar("_Config", bound="Config")
 
-T = TypeVar("T")  #: Item type
+_Value = TypeVar("_Value")
 
 
 class CLHandler(ABC):
@@ -66,7 +91,7 @@ class CLSpecialAction(CLHandler):
     A handler that sets the special action
     """
 
-    special_action: SpecialAction
+    special_action: SpecialAction  #: Special action to set
 
     def handle(self, args: Sequence[str], state: State) -> Tuple[Sequence[str], Optional[Err]]:
         if state.special_action is not None:
@@ -84,6 +109,7 @@ class CLInserter(CLHandler):
     Handler that expands a flag into a sequence of args inserted into the command line to be parsed
     """
 
+    #: Arguments inserted in the command-line
     inserted_args: Sequence[str]
 
     def handle(self, args: Sequence[str], state: State) -> Tuple[Sequence[str], Optional[Err]]:
@@ -92,7 +118,7 @@ class CLInserter(CLHandler):
 
 
 @dataclass(frozen=True)
-class CLParam(CLHandler, Generic[T]):
+class CLParam(CLHandler, Generic[_Value]):
     """
     Parameter handler
 
@@ -100,9 +126,10 @@ class CLParam(CLHandler, Generic[T]):
     corresponding sequence of instances
     """
 
-    param: Param[T]
+    #: Parameter to handle
+    param: Param[_Value]
 
-    def action(self, value: T, state: State) -> Optional[Err]:
+    def action(self, value: _Value, state: State) -> Optional[Err]:
         """
         A method called on the successful parse of a value
 
@@ -119,16 +146,19 @@ class CLParam(CLHandler, Generic[T]):
 
     def handle(self, args: Sequence[str], state: State) -> Tuple[Sequence[str], Optional[Err]]:
         if args:
-            res = self.param.param_type.parse(args[0])
+            res = self.param.parser.parse(args[0])
             if isinstance(res, Err):
-                return (args[1:], res)
+                return (args[1:], res.in_context(param=self.param.name))
             else:
                 assert self.param.name is not None, "Names are assigned after initialization"
-                err = self.action(res, state)
+                err = in_context(self.action(res, state), param=self.param.name)
                 state.instances[self.param.name] = [*state.instances[self.param.name], res]
                 return (args[1:], err)
         else:
-            return (args, Err.make("Expected value, but no argument present"))
+            return (
+                args,
+                Err.make("Expected value, but no argument present", param=self.param.name),
+            )
 
 
 @dataclass(frozen=True)
@@ -166,11 +196,9 @@ class CLPos(CLHandler):
         Returns:
             Handler
         """
+        assert all([p.positional is not None for p in seq]), "All parameters should be positional"
         assert all(
-            [p.positional.is_positional() for p in seq]
-        ), "All parameters should be positional"
-        assert all(
-            [not p.positional.should_be_last() for p in seq[:-1]]
+            [not p.positional.should_be_last() for p in seq[:-1] if p.positional is not None]
         ), "Positional parameters with a variable number of arguments should be last"
         l = list(seq)  # makes a mutable copy
         return CLPos(l)
@@ -182,9 +210,9 @@ class CLPos(CLHandler):
             return (args[1:], Err.make(f"Unknown argument {args[0]}"))
         p = self.pos[0]
         assert p.name is not None
-        res = p.param_type.parse(args[0])
+        res = p.parser.parse(args[0])
         if isinstance(res, Err):
-            return (args[1:], res)
+            return (args[1:], in_context(res, param=p.name))
         else:
             state.append(p.name, res)
             if p.positional == Positional.ONCE:
@@ -241,10 +269,15 @@ class KVHandler(ABC):
 
 
 @dataclass(frozen=True)
-class KVParam(KVHandler, Generic[T]):
-    param: Param[T]
+class KVParam(KVHandler, Generic[_Value]):
+    """
+    Handler for the value following a key corresponding to a parameter
+    """
 
-    def action(self, value: T, state: State) -> Optional[Err]:
+    #: Parameter to handle
+    param: Param[_Value]
+
+    def action(self, value: _Value, state: State) -> Optional[Err]:
         """
         A method called on the successful parse of a value
 
@@ -260,18 +293,22 @@ class KVParam(KVHandler, Generic[T]):
         return None
 
     def handle(self, value: str, state: State) -> Optional[Err]:
-        res = self.param.param_type.parse(value)
+        res = self.param.parser.parse(value)
         if isinstance(res, Err):
             return res
         else:
             assert self.param.name is not None
             err = self.action(res, state)
             state.instances[self.param.name] = [*state.instances[self.param.name], res]
-            return err
+            return in_context(err, param=self.param.name)
 
 
 @dataclass(frozen=True)
 class KVConfigParam(KVParam[Sequence[Path]]):
+    """
+    Handler for the configuration file value following a key corresponding to a parameter
+    """
+
     def action(self, value: Sequence[Path], state: State) -> Optional[Err]:
         state.config_files_to_process.extend(value)
         return None
@@ -328,7 +365,7 @@ class State:
         for p in params:
             assert p.name is not None, "Arguments have names after initialization"
             if p.default_value is not None:
-                res = p.param_type.parse(p.default_value)
+                res = p.parser.parse(p.default_value)
                 if isinstance(res, Err):
                     raise ValueError(f"Invalid default {p.default_value} for parameter {p.name}")
                 instances[p.name] = [res]
@@ -346,7 +383,64 @@ class IniProcessor:
     section_strict: Mapping[str, bool]  #: Sections and their strictness
     kv_handlers: Mapping[str, KVHandler]  #: Handler for key/value pairs
 
-    def process(self, ini_path: Path, state: State) -> Optional[Err]:
+    def _process(self, parser: ConfigParser, state: State) -> Sequence[Err]:
+        """
+        Processes a filled ConfigParser
+
+        Args:
+            parser: INI data to parse
+            state: Mutable state to update
+
+        Returns:
+            Errors that occurred, if any
+        """
+        errors: List[Err] = []
+        try:
+            for section_name in parser.sections():
+                if section_name in self.section_strict:
+                    for key, value in parser[section_name].items():
+                        err: Optional[Err] = None
+                        if key in self.kv_handlers:
+                            res = self.kv_handlers[key].handle(value, state)
+                            if isinstance(res, Err):
+                                err = res
+                        else:
+                            if self.section_strict[section_name]:
+                                err = Err.make(f"Unknown key {key}")
+                        if err is not None:
+                            errors.append(err.in_context(ini_section=section_name))
+        except configparser.Error as e:
+            errors.append(Err.make(f"Parse error"))
+        except IOError as e:
+            errors.append(Err.make(f"IO Error"))
+        return errors
+
+    def process_string(self, ini_contents: str, state: State) -> Optional[Err]:
+        """
+        Processes a configuration file given as a string
+
+        Args:
+            ini_contents: Contents of the INI file
+            state: Mutable state to update
+
+        Returns:
+            An optional error
+        """
+        errors: List[Err] = []
+        parser = ConfigParser()
+        try:
+            parser.read_string(ini_contents)
+            errors.extend(self._process(parser, state))
+        except configparser.Error as e:
+            errors.append(Err.make(f"Parse error"))
+        except IOError as e:
+            errors.append(Err.make(f"IO Error"))
+        if errors:
+            return Err.collect(*errors)
+        else:
+            return None
+
+    def process(self, ini_file_path: Path, state: State) -> Optional[Err]:
         """
         Processes a configuration file
 
@@ -358,27 +452,15 @@ class IniProcessor:
             An optional error
         """
         errors: List[Err] = []
-        if not ini_path.exists():
-            return Err.make(f"Config file {ini_path} does not exist")
-        if not ini_path.is_file():
-            return Err.make(f"Path {ini_path} is not a file")
+        if not ini_file_path.exists():
+            return Err.make(f"Config file {ini_file_path} does not exist")
+        if not ini_file_path.is_file():
+            return Err.make(f"Path {ini_file_path} is not a file")
         parser = ConfigParser()
         try:
-            with open(ini_path, "r") as file:
+            with open(ini_file_path, "r") as file:
                 parser.read_file(file)
-                for section_name in parser.sections():
-                    if section_name in self.section_strict:
-                        for key, value in parser[section_name].items():
-                            err: Optional[Err] = None
-                            if key in self.kv_handlers:
-                                res = self.kv_handlers[key].handle(value, state)
-                                if isinstance(res, Err):
-                                    err = res
-                            else:
-                                if self.section_strict[section_name]:
-                                    err = Err.make(f"Unknown key {key}")
-                            if err is not None:
-                                errors.append(err.in_context(ini_section=section_name))
+                errors.extend(self._process(parser, state))
         except configparser.Error as e:
             errors.append(Err.make(f"Parse error"))
         except IOError as e:
@@ -390,7 +472,7 @@ class IniProcessor:
 
 
 @dataclass
-class ProcessorFactory(Generic[C]):
+class ProcessorFactory(Generic[_Config]):
     """
     Describes a processor in construction
 
@@ -427,10 +509,10 @@ class ProcessorFactory(Generic[C]):
     #: List of positional arguments
     cl_positionals: List[Param[Any]]
 
-    validators: List[Callable[[C], Optional[Err]]]
+    validators: List[Callable[[_Config], Optional[Err]]]
 
     @staticmethod
-    def make(config_type: Type[C]) -> ProcessorFactory[C]:
+    def make(config_type: Type[_Config]) -> ProcessorFactory[_Config]:
         """
         Constructs an empty processor factory
 
@@ -473,13 +555,13 @@ class ProcessorFactory(Generic[C]):
 
 
 @dataclass(frozen=True)
-class Processor(Generic[C]):
+class Processor(Generic[_Config]):
     """
     Configuration processor
     """
 
     #: Configuration to parse
-    config_type: Type[C]
+    config_type: Type[_Config]
 
     #: Completed argument parser, used only for documentation purposes (CLI and Sphinx)
     argument_parser: argparse.ArgumentParser
@@ -491,15 +573,15 @@ class Processor(Generic[C]):
     ini_processor: IniProcessor
 
     #: Command line arguments handler
-    cl_handler: CLHandler
+    cl_handler: CLStdHandler
 
     #: Dictionnary of parameters by field name
     params_by_name: Mapping[str, Param[Any]]
 
-    validators: Sequence[Callable[[C], Optional[Err]]]
+    validators: Sequence[Callable[[_Config], Optional[Err]]]
 
     @staticmethod
-    def process_fields(config_type: Type[C]) -> Sequence[Arg]:
+    def process_fields(config_type: Type[_Config]) -> Sequence[Arg]:
         """
         Returns a sequence of the arguments present in a configuration, with updated data
 
@@ -510,7 +592,7 @@ class Processor(Generic[C]):
             Sequence of arguments
         """
         args: List[Arg] = []
-        docs: ClassDoc[C] = ClassDoc.make(config_type)
+        docs: ClassDoc[_Config] = ClassDoc.make(config_type)
         th = get_type_hints(config_type, include_extras=True)
         for name, typ in th.items():
             arg: Optional[Arg] = None
@@ -535,8 +617,8 @@ class Processor(Generic[C]):
 
     @staticmethod
     def make(
-        config_type: Type[C],
-    ) -> Processor[C]:
+        config_type: Type[_Config],
+    ) -> Processor[_Config]:
         """
         Creates the processor corresponding to a configuration
         """
@@ -544,6 +626,9 @@ class Processor(Generic[C]):
         pf = ProcessorFactory.make(config_type)
         for arg in Processor.process_fields(config_type):
             arg.update_processor(pf)
+
+        # if these flags are no longer provided by default, update the overview concept notebook
+        # in the documentation
         pf.cl_flag_handlers["-h"] = CLSpecialAction(SpecialAction.HELP)
         pf.cl_flag_handlers["--help"] = CLSpecialAction(SpecialAction.HELP)
 
@@ -557,7 +642,7 @@ class Processor(Generic[C]):
             validators=pf.validators,
         )
 
-    def process_config(self, cwd: Path, state: State) -> Optional[Err]:
+    def _process_config(self, cwd: Path, state: State) -> Optional[Err]:
         """
         Processes configuration files if such processing was requested by a handler
 
@@ -577,12 +662,104 @@ class Processor(Generic[C]):
                 errors.append(err.in_context(ini_file=p))
         return Err.collect(*errors)
 
+    def process_ini_contents(self, ini_contents: str) -> Res[_Config]:
+        """
+        Processes the configuration in INI format contained in a string
+
+        Args:
+            ini_contents: Multiline string containing information in INI format
+
+        Returns:
+            The parsed configuration or an error
+        """
+        state = self._state_with_default_values()
+        err = self.ini_processor.process_string(ini_contents, state)
+        if err is not None:
+            return err
+        return self._finish_processing_state(state)
+
+    def process_ini_file(self, ini_file_path: Path) -> Res[_Config]:
+        """
+        Processes the configuration contained in an INI file
+
+        Args:
+            ini_file_path: Path to the file to parse
+
+        Returns:
+            The parsed configuration or an error
+        """
+        state = self._state_with_default_values()
+        err = self.ini_processor.process(ini_file_path, state)
+        if err is not None:
+            return err
+        return self._finish_processing_state(state)
+
     def process(
         self,
         cwd: Path,
         args: Sequence[str],
         env: Mapping[str, str],
-    ) -> Result[Union[C, SpecialAction]]:
+    ) -> Res[Union[_Config, SpecialAction]]:
+        """
+        Processes command-line arguments (deprecated)
+
+        See also: :meth:`.process_command_line`
+        """
+        warnings.warn(
+            "process has been deprecated, use process_command_line instead",
+            DeprecationWarning,
+            stacklevel=2,
+        )
+        return self.process_command_line(cwd, args, env)
+
+    def _finish_processing_state(self, state: State) -> Res[_Config]:
+        """
+        Finishes the processing of the state
+
+        This method performs:
+
+        - the collection of parameter values
+        - the final validation using validation methods
+
+        Args:
+            state: State after parsing all elements, must have :attr:`.State.special_action` set
+                   to :data:`None`
+
+        Returns:
+            The parsed configuration or an error
+        """
+        assert state.special_action is None
+        errors: List[Err] = []
+        collected: Dict[str, Any] = {}
+        for name, param in self.params_by_name.items():
+            instances = state.instances[name]
+            res = param.collector.collect(instances)
+            if isinstance(res, Err):
+                errors.append(res.in_context(param=name))
+            else:
+                collected[name] = res
+
+        if errors:
+            return Err.collect1(*errors)
+        c: _Config = self.config_type(**collected)
+        validation_error: Optional[Err] = Err.collect(*[f(c) for f in self.validators])
+        if validation_error is not None:
+            return validation_error
+        else:
+            return c
+
+    def _state_with_default_values(self) -> State:
+        """
+        Returns a new state instance with default values populated
+        """
+        return State.make(self.params_by_name.values())
+
+    def process_command_line(
+        self,
+        cwd: Path,
+        args: Sequence[str],
+        env: Mapping[str, str],
+    ) -> Res[Union[_Config, SpecialAction]]:
         """
         Processes command-line arguments, configuration files and environment variables
 
@@ -595,7 +772,7 @@ class Processor(Generic[C]):
             Either a parsed configuration, a special action to execute, or (a list of) errors
         """
         errors: List[Err] = []
-        state = State.make(self.params_by_name.values())
+        state = self._state_with_default_values()
         # process environment variables
         for key, value in env.items():
             handler = self.env_handlers.get(key)
@@ -603,7 +780,7 @@ class Processor(Generic[C]):
                 err = handler.handle(value, state)
                 if err is not None:
                     errors.append(err.in_context(environment_variable=key))
-            err = self.process_config(cwd, state)
+            err = self._process_config(cwd, state)
             if err is not None:
                 errors.append(err.in_context(environment_variable=key))
         # process command line arguments
@@ -612,7 +789,7 @@ class Processor(Generic[C]):
             rest_args, err = self.cl_handler.handle(rest_args, state)
             if err is not None:
                 errors.append(err)
-            err = self.process_config(cwd, state)
+            err = self._process_config(cwd, state)
             if err is not None:
                 errors.append(err)
 
@@ -620,25 +797,6 @@ class Processor(Generic[C]):
             return state.special_action
 
         if errors:
-            return Err.collect_non_empty(*errors)
-        collected: Dict[str, Any] = {}
-        for name, param in self.params_by_name.items():
-            instances = state.instances[name]
-            res = param.collector.collect(instances)
-            if is_value(res) and param.validator is not None:
-                res1 = param.validator(res)
-                if res1 is not None:
-                    res = res1
-            if is_err(res):
-                errors.append(res.in_context(param=name))
-            else:
-                collected[name] = res
+            return Err.collect1(*errors)
 
-        if errors:
-            return Err.collect_non_empty(*errors)
-        c: C = self.config_type(**collected)
-        validation_error: Optional[Err] = Err.collect_optional(*[f(c) for f in self.validators])
-        if validation_error is not None:
-            return validation_error
-        else:
-            return c
+        return self._finish_processing_state(state)
